/*
 * $Id: label.c,v 1.2 2004/01/19 13:08:20 chan12 Exp $
 *
 * Copyright (C) 1999-2004 Tak-Shing Chan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

/* Input buffer */
#define LINES   32
#define BUFSIZE 128
static char     buffer[LINES][BUFSIZE];
static int      copies;

/* Read records from input file */
/* Returns: number of lines read */
static int
readnext(FILE *fp, int rows)
{
    static int  trunc = 0;
    int         i = 0, j, k, len;
    const char  postnet[] = { 24, 3, 5, 6, 9, 10, 12, 17, 18, 20 },
                *cbc = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

    /* Input loop */
    copies = 1;
    while (fgets(buffer[i], BUFSIZE - 1, fp)) {

        /* Truncate lines into buffer, and discard newlines */
        len = strlen(buffer[i]) - 1;
        if (buffer[i][len] == '\n') {
            if (trunc) {
                trunc = 0;
                continue;
            } else
                buffer[i][len] = 0;
        } else {
            if (trunc)
                continue;
            else
                trunc = 1;
        }

        /* Mark buffer as plain text */
        buffer[i][BUFSIZE - 1] = 'S';

        /* Handle requests */
        if (!strcmp(buffer[i], "####")) {
            if (i)
                return i;
            else
                continue;
        }
        if (len > 2 && buffer[i][0] == '.') {

            /* Number of copies */
            if (buffer[i][1] == 'c' && buffer[i][2] == 'p') {
                if (sscanf(buffer[i] + 3, "%u", &copies) != 1 ||
                    copies < 1)
                    copies = 1;
                continue;
            }

            /* POSTNET barcodes */
            else if (buffer[i][1] == 'p' && buffer[i][2] == 'n') {
                char code, bar[12];
                int sum = 0, count = 0;
                for (j = 3; count <= 10 && j < len; j++)
                    if (isdigit((unsigned char) (code =
                        buffer[i][j])))
                        sum += (bar[count++] = code - '0');

                /* Compute check digit */
                bar[count] = (10 - sum % 10) % 10;

                /* Convert string to POSTNET binary */
                buffer[i][0] = '1';
                for (j = 0; j <= count; j++) {
                    code = postnet[(int) bar[j]];
                    for (k = 5; k; code >>= 1)
                        buffer[i][j * 5 + k--] = '0' + (code & 1);
                }
                buffer[i][j * 5 + 1] = '1';
                buffer[i][j * 5 + 2] = 0;

                /* Mark buffer as POSTNET code */
                buffer[i][BUFSIZE - 1] = 'P';
            }

            /* Royal Mail CBC */
            else if (buffer[i][1] == 'r' && buffer[i][2] == 'm') {
                char rbar[12], cbar[12];
                int rsum = 0, csum = 0, count = 0;
                for (j = 3; count <= 10 && j < len; j++) {
                    char *code;
                    if ((code = strchr(cbc, buffer[i][j]))) {
                        div_t d = div(code - cbc, 6);
                        rsum += (rbar[count] = d.quot + 1);
                        csum += (cbar[count++] = d.rem + 1);
                    }
                }

                /* Compute check digit */
                rbar[count] = rsum % 6;
                cbar[count] = csum % 6;

                /* Convert string to CBC */
                buffer[i][0] = '2';
                for (j = 0; j <= count; j++) {
                    buffer[i][j * 4 + 1] = '0' +
                        ((rbar[j] & 4) >> 1) + ((cbar[j] & 4) >> 2);
                    buffer[i][j * 4 + 2] = '0' +
                        (rbar[j] & 2) + ((cbar[j] & 2) >> 1);
                    buffer[i][j * 4 + 3] = '0' +
                        ((rbar[j] & 1) << 1) + (cbar[j] & 1);

                    /* Parity bits */
                    buffer[i][j * 4 + 4] = '0' +
                        (rbar[j] == 1 || rbar[j] == 2 || rbar[j] == 4) * 2 +
                        (cbar[j] == 1 || cbar[j] == 2 || cbar[j] == 4);
                }
                buffer[i][j * 4 + 1] = '3';
                buffer[i][j * 4 + 2] = 0;

                /* Mark buffer as CBC */
                buffer[i][BUFSIZE - 1] = 'C';
            }
        }
        /* Silently discards extra rows */
        if (i < rows) i++;
    }

    /* EOF */
    return i - (i == rows);
}

int
main(int argc, char *argv[])
{
    int             i, opened = 0, flag = 0, bolding = 0,
                    font = 0, indent = 12, width = 288,
                    height = 96, hsize = 612, vsize = 792,
                    point = 12, hmargin, vmargin, num, page = 1,
                    rows, hpos, vpos;
    FILE            *fp;
    time_t          now;
    char            *roman[] = { "Helvetica", "Courier" },
                    *bold[] = { "Helvetica-Bold", "Courier-Bold" };

    /* Process arguments */
    for (i = 1; i < argc; i++) {
        if (argv[i][0] != '-') {
            /* Bail out if input file already opened */
            if (opened) {
                flag++;
            } else {
                /* Otherwise, open it now */
                if (!(fp = fopen(argv[i], "r"))) {
                    perror(argv[i]);
                    exit(EXIT_FAILURE);
                }
                opened = i;
            }
        } else {
            switch (argv[i][1]) {
            case 'b':
                bolding = 1;
                break;
            case 'c':
                font = 1;
                break;
            case 'i':
                if (sscanf(&argv[i][2], "%d", &indent) != 1)
                    flag++;
                break;
            case 'l':
                if (sscanf(&argv[i][2], "%dx%d", &width, &height)
                    != 2 || width < 1 || height < 1)
                    flag++;
                break;
            case 'p':
                if (sscanf(&argv[i][2], "%dx%d", &hsize, &vsize)
                    != 2 || hsize < 1 || vsize < 1)
                    flag++;
                break;
            case 's':
                if (sscanf(&argv[i][2], "%d", &point)
                    != 1 || point < 10 || point > 12)
                    flag++;
                break;
            default:
                flag++;
            }
        }

        /* Illegal options? */
        if (flag) {
            fprintf(stderr, "Usage: %s [-b] [-c] [-iINDENT] [-lW"
"xH] [-pWxH] [-s10|11|12] [FILE]\n", argv[0]);
            exit(EXIT_FAILURE);
        }
    }

    /* Use stdin if no file is specified */
    if (!opened)
        fp = stdin;

    /* PostScript Level 2 output (DSC 3.0) */
    puts("%!PS-Adobe-3.0");
    printf("%%%%Title: %s\n", opened ? argv[opened] : "stdin");
    puts("%%Creator: Label");
    time(&now);
    printf("%%%%CreationDate: %s", ctime(&now));
    puts("%%Pages: (atend)");
    puts("%%Orientation: Portrait");
    printf("%%%%DocumentNeededResources: font %s", roman[font]);
    if (bolding)
        printf(" %s\n", bold[font]);
    else
        putchar('\n');
    puts("%%DocumentData: Clean7Bit");
    puts("%%LanguageLevel: 2");
    puts("%%EndComments");
    puts("%%BeginSetup");
    printf("%%%%IncludeResource: font %s\n", roman[font]);
    if (bolding)
        printf("%%%%IncludeResource: font %s\n", bold[font]);
    hmargin = (hsize / (hsize / width) - width) / 2;
    vmargin = (vsize - ((vsize - 72) / height) * height) / 2;
    printf("/I { initclip %d %d %d rectclip ", vmargin, width -
        indent, vsize - 2 * vmargin);
    if (bolding) {
        printf("/%s %d selectfont } bind def\n", bold[font],
            point);
        printf("/R { /%s %d selectfont } bind def\n",
            roman[font], point);
    } else {
        printf("/%s %d selectfont } bind def\n", roman[font],
            point);
    }
    puts("/S { moveto show } bind def");
    puts("/F { 3.3 9 rmoveto 0 -9 rlineto } bind def");
    puts("/H { 3.3 3.6 rmoveto 0 -3.6 rlineto } bind def");
    puts("/P { moveto { 49 eq { F } { H } ifelse } forall stroke"
         " } bind def");
    puts("/T { 3.3 8.93 rmoveto 0 -3.6 rlineto 0 -5.33 rmoveto }"
         " bind def");
    puts("/D { 3.3 8.93 rmoveto 0 -8.93 rlineto 0 } bind def");
    puts("/A { 3.3 14.26 rmoveto 0 -8.93 rlineto 0 -5.33 rmoveto"
         " } bind def");
    puts("/B { 3.3 14.26 rmoveto 0 -14.26 rlineto } bind def");
    puts("/C { moveto { [(T) (D) (A) (B)] exch 48 sub get cvx ex"
         "ec } forall stroke } bind def");
    puts("%%EndSetup");

    /* Main loop */
    if ((rows = height / (point + 1)) >= LINES)
        rows = LINES - 1;
    for (num = readnext(fp, rows); num; page++) {
        printf("%%%%Page: %d %d\n", page, page);
        printf("%%%%PageResources: font %s", roman[font]);
        if (bolding)
            printf(" %s\n", bold[font]);
        else
            putchar('\n');
        puts("%%BeginPageSetup");
        puts("/pagelevel save def");
        puts("1.44 setlinewidth");
        puts("%%EndPageSetup");
        for (vpos = vsize - vmargin; num && vpos >= height +
            vmargin; vpos -= height) {
            for (hpos = hmargin; num && hpos <= hsize - hmargin -
                width; hpos += 2 * hmargin + width) {
                for (i = 0; i < num; i++) {
                    switch (i) {
                    case 0:
                        printf("%d I ", hpos);
                        break;
                    case 1:
                        if (bolding)
                            printf("R ");
                    }
                    printf("(%s) %d %d %c\n", buffer[i], hpos +
                        indent, vpos + ((point + 1) * (num - 2 *
                        i - 2) - height) / 2, buffer[i][BUFSIZE -
                        1]);
                }
                if (!--copies)
                    num = feof(fp) ? 0 : readnext(fp, rows);
            }
        }
        puts("pagelevel restore");
        puts("showpage");
    }

    /* Trailer */
    puts("%%Trailer");
    printf("%%%%Pages: %d\n", page - 1);
    puts("%%EOF");
    if (opened)
        fclose(fp);
    return EXIT_SUCCESS;
}
